# pylint: disable=missing-docstring

import contextlib
import io
import json
import os
import pathlib
import tempfile
import unittest
import warnings

import tests.common

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    # NOTE (mristin, 2022-04-08):
    # We need to disable warnings. Jsonschema package at the latest version (4.4.0) has
    # a problem with JSON schema draft 2019-09 and crashes with an recursion error,
    # see: https://github.com/python-jsonschema/jsonschema/issues/847.
    #
    # We revert back to jsonschema 3.2.0, which can not handle 2019-09, but still seems
    # to validate correctly our examples.
    import jsonschema

import aas_core_codegen.main
import aas_core_codegen.jsonschema.main


class Test_against_recorded(unittest.TestCase):
    _REPO_DIR = pathlib.Path(os.path.realpath(__file__)).parent.parent.parent.parent
    PARENT_CASE_DIR = _REPO_DIR / "dev" / "test_data" / "jsonschema" / "test_main"

    def test_against_meta_models(self) -> None:
        test_cases = tests.common.test_cases_from_base_case_dir(
            base_case_dir=Test_against_recorded.PARENT_CASE_DIR
        ) + tests.common.test_cases_from_real_world_models(
            base_case_dir=Test_against_recorded.PARENT_CASE_DIR,
            real_meta_model_paths=tests.common.REAL_META_MODEL_PATHS,
        )

        for test_case in test_cases:
            snippets_dir = test_case.case_dir / "input/snippets"
            assert snippets_dir.exists() and snippets_dir.is_dir(), snippets_dir

            expected_output_dir = test_case.case_dir / "expected_output"

            with contextlib.ExitStack() as exit_stack:
                if tests.common.RERECORD:
                    output_dir = expected_output_dir
                    expected_output_dir.mkdir(exist_ok=True, parents=True)
                else:
                    assert (
                        expected_output_dir.exists() and expected_output_dir.is_dir()
                    ), expected_output_dir

                    # pylint: disable=consider-using-with
                    tmp_dir = tempfile.TemporaryDirectory()
                    exit_stack.push(tmp_dir)
                    output_dir = pathlib.Path(tmp_dir.name)

                params = aas_core_codegen.main.Parameters(
                    model_path=test_case.model_path,
                    target=aas_core_codegen.main.Target.JSONSCHEMA,
                    snippets_dir=snippets_dir,
                    output_dir=output_dir,
                )

                stdout = io.StringIO()
                stderr = io.StringIO()

                return_code = aas_core_codegen.main.execute(
                    params=params, stdout=stdout, stderr=stderr
                )

                if stderr.getvalue() != "":
                    raise AssertionError(
                        f"Expected no stderr on valid models, but got:\n"
                        f"{stderr.getvalue()}"
                    )

                self.assertEqual(
                    0, return_code, "Expected 0 return code on valid models"
                )

                stdout_pth = expected_output_dir / "stdout.txt"
                normalized_stdout = stdout.getvalue().replace(
                    str(output_dir), "<output dir>"
                )

                if tests.common.RERECORD:
                    stdout_pth.write_text(normalized_stdout, encoding="utf-8")
                else:
                    self.assertEqual(
                        normalized_stdout,
                        stdout_pth.read_text(encoding="utf-8"),
                        stdout_pth,
                    )

                for relevant_rel_pth in [
                    pathlib.Path("schema.json"),
                ]:
                    expected_pth = expected_output_dir / relevant_rel_pth
                    output_pth = output_dir / relevant_rel_pth

                    if not output_pth.exists():
                        raise FileNotFoundError(
                            f"The output file is missing: {output_pth}"
                        )

                    if tests.common.RERECORD:
                        expected_pth.write_text(
                            output_pth.read_text(encoding="utf-8"), encoding="utf-8"
                        )
                    else:
                        self.assertEqual(
                            expected_pth.read_text(encoding="utf-8"),
                            output_pth.read_text(encoding="utf-8"),
                            f"The files {expected_pth} and {output_pth} do not match.",
                        )

    def test_on_examples(self) -> None:
        assert (
            Test_against_recorded.PARENT_CASE_DIR.exists()
            and Test_against_recorded.PARENT_CASE_DIR.is_dir()
        ), f"{Test_against_recorded.PARENT_CASE_DIR=}"

        for case_dir in Test_against_recorded.PARENT_CASE_DIR.iterdir():
            assert case_dir.is_dir(), case_dir

            schema_pth = case_dir / "expected_output" / "schema.json"

            with schema_pth.open("rt", encoding="utf-8") as fid:
                schema = json.load(fid)

            for data_pth in sorted(
                (case_dir / "examples" / "expected").glob("**/*.json")
            ):
                with data_pth.open("rt", encoding="utf-8") as fid:
                    instance = json.load(fid)

                try:
                    jsonschema.validate(instance=instance, schema=schema)
                except jsonschema.ValidationError as err:
                    raise AssertionError(
                        f"Failed to validate {data_pth} against {schema_pth}"
                    ) from err


class Test_pattern_transpilation(unittest.TestCase):
    def test_unescaped_above_ascii_character_in_bmp(self) -> None:
        pattern = aas_core_codegen.jsonschema.main.fix_pattern_for_utf16("\ud7ff")
        self.assertEqual("\ud7ff", pattern)

    def test_escaped_above_ascii_character_in_bmp(self) -> None:
        pattern = aas_core_codegen.jsonschema.main.fix_pattern_for_utf16("\\ud7ff")
        self.assertEqual("\\ud7ff", pattern)

    def test_unescaped_range_in_bmp(self) -> None:
        pattern = aas_core_codegen.jsonschema.main.fix_pattern_for_utf16(
            "[\x20-\uD7FF]"
        )
        self.assertEqual("[\x20-\uD7FF]", pattern)

    def test_escaped_range_in_bmp(self) -> None:
        pattern = aas_core_codegen.jsonschema.main.fix_pattern_for_utf16(
            "[\\x20-\\ud7ff]"
        )
        self.assertEqual("[\\x20-\\ud7ff]", pattern)

    def test_escaped_range_in_bmp_always_lowercase(self) -> None:
        pattern = aas_core_codegen.jsonschema.main.fix_pattern_for_utf16(
            "[\\x20-\\uD7FF]"
        )
        self.assertEqual("[\\x20-\\ud7ff]", pattern)

    def test_unescaped_above_bmp(self) -> None:
        pattern = aas_core_codegen.jsonschema.main.fix_pattern_for_utf16(
            "[\U00010000-\U0010FFFF]"
        )
        self.assertEqual(
            "(\\ud800[\\udc00-\\udfff]|[\\ud801-\\udbfe][\\udc00-\\udfff]"
            "|\\udbff[\\udc00-\\udfff])",
            pattern,
        )

    def test_escaped_above_bmp(self) -> None:
        pattern = aas_core_codegen.jsonschema.main.fix_pattern_for_utf16(
            "[\\U00010000-\\U0010FFFF]"
        )
        self.assertEqual(
            "(\\ud800[\\udc00-\\udfff]|[\\ud801-\\udbfe][\\udc00-\\udfff]"
            "|\\udbff[\\udc00-\\udfff])",
            pattern,
        )

    def test_unescaped_special_ascii_characters(self) -> None:
        pattern = aas_core_codegen.jsonschema.main.fix_pattern_for_utf16("[\t\n\r]")
        self.assertEqual("[\\t\\n\\r]", pattern)

    def test_escaped_special_ascii_characters(self) -> None:
        pattern = aas_core_codegen.jsonschema.main.fix_pattern_for_utf16("[\\t\\n\\r]")
        self.assertEqual("[\\t\\n\\r]", pattern)

    def test_on_XML_string_pattern(self) -> None:
        pattern = aas_core_codegen.jsonschema.main.fix_pattern_for_utf16(
            r"^[\x09\x0A\x0D\x20-\uD7FF\uE000-\uFFFD\U00010000-\U0010FFFF]*$"
        )
        self.assertEqual(
            "^([\\x09\\x0a\\x0d\\x20-\\ud7ff\\ue000-\\ufffd]|\\ud800[\\udc00-\\udfff]"
            "|[\\ud801-\\udbfe][\\udc00-\\udfff]|\\udbff[\\udc00-\\udfff])*$",
            pattern,
        )

        # NOTE (mristin, 2024-05-08):
        # We also test for the JSON representation to test against a possible bug
        # reported in:
        # https://github.com/admin-shell-io/aas-specs/pull/426 and
        # https://github.com/aas-core-works/aas-core-codegen/issues/485

        pattern_json = json.dumps(pattern)
        self.assertEqual(
            # NOTE (mristin, 2024-05-08):
            # Mind the ``r`` modifier for the string literals!
            r'"^([\\x09\\x0a\\x0d\\x20-\\ud7ff\\ue000-\\ufffd]'
            r"|\\ud800[\\udc00-\\udfff]|[\\ud801-\\udbfe][\\udc00-\\udfff]"
            r'|\\udbff[\\udc00-\\udfff])*$"',
            pattern_json,
        )


if __name__ == "__main__":
    unittest.main()



# We explicitly test here that the parsed order of the contracts and snapshots is
# reversed w.r.t. the source code. This reversion is due to how Python applies
# decorators: from bottom up.


class Something:
    @require(lambda x: x > 0)
    @require(lambda y: y > 0)
    @snapshot(lambda x: x, name="double_x")
    @ensure(lambda x, result: x > result)
    @ensure(lambda y, result: y > result)
    def do_something(self, x: int, y: int) -> int:
        pass


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



# pylint: disable=missing-docstring

import contextlib
import io
import os
import pathlib
import tempfile
import unittest

import aas_core_codegen.main

import tests.common

REPO_DIR = pathlib.Path(os.path.realpath(__file__)).parent.parent.parent.parent


class Test_against_recorded(unittest.TestCase):
    def test_against_expected_meta_models(self) -> None:
        base_case_dir = (
            REPO_DIR / "dev" / "test_data" / "proto" / "test_main" / "expected"
        )

        test_cases = tests.common.test_cases_from_base_case_dir(
            base_case_dir=base_case_dir
        ) + tests.common.test_cases_from_real_world_models(
            base_case_dir=base_case_dir,
            real_meta_model_paths=tests.common.REAL_META_MODEL_PATHS,
        )

        for test_case in test_cases:
            snippets_dir = test_case.case_dir / "input/snippets"
            assert snippets_dir.exists() and snippets_dir.is_dir(), snippets_dir

            expected_output_dir = test_case.case_dir / "expected_output"

            with contextlib.ExitStack() as exit_stack:
                if tests.common.RERECORD:
                    output_dir = expected_output_dir
                    expected_output_dir.mkdir(exist_ok=True, parents=True)
                else:
                    assert (
                        expected_output_dir.exists() and expected_output_dir.is_dir()
                    ), expected_output_dir

                    # pylint: disable=consider-using-with
                    tmp_dir = tempfile.TemporaryDirectory()
                    exit_stack.push(tmp_dir)
                    output_dir = pathlib.Path(tmp_dir.name)

                params = aas_core_codegen.main.Parameters(
                    model_path=test_case.model_path,
                    target=aas_core_codegen.main.Target.PROTOBUF,
                    snippets_dir=snippets_dir,
                    output_dir=output_dir,
                )

                stdout = io.StringIO()
                stderr = io.StringIO()

                return_code = aas_core_codegen.main.execute(
                    params=params, stdout=stdout, stderr=stderr
                )

                if stderr.getvalue() != "":
                    raise AssertionError(
                        f"Expected no stderr on valid models, but got "
                        f"for {test_case.model_path}:\n"
                        f"{stderr.getvalue()}"
                    )

                self.assertEqual(
                    0, return_code, "Expected 0 return code on valid models"
                )

                stdout_pth = expected_output_dir / "stdout.txt"
                normalized_stdout = stdout.getvalue().replace(
                    str(output_dir), "<output dir>"
                )

                if tests.common.RERECORD:
                    stdout_pth.write_text(normalized_stdout, encoding="utf-8")
                else:
                    self.assertEqual(
                        normalized_stdout,
                        stdout_pth.read_text(encoding="utf-8"),
                        stdout_pth,
                    )

                for relevant_rel_pth in [pathlib.Path("types.proto")]:
                    expected_pth = expected_output_dir / relevant_rel_pth
                    output_pth = output_dir / relevant_rel_pth

                    if not output_pth.exists():
                        raise FileNotFoundError(
                            f"The output file is missing: {output_pth}"
                        )

                    if tests.common.RERECORD:
                        expected_pth.write_text(
                            data=output_pth.read_text(encoding="utf-8"),
                            encoding="utf-8",
                        )
                    else:
                        self.assertEqual(
                            expected_pth.read_text(encoding="utf-8"),
                            output_pth.read_text(encoding="utf-8"),
                            f"The files {expected_pth} and {output_pth} do not match.",
                        )


if __name__ == "__main__":
    unittest.main()



# pylint: disable=missing-docstring
import os
import pathlib
import unittest

from aas_core_codegen import intermediate
from aas_core_codegen.common import LinenoColumner, Error
from aas_core_codegen.csharp import (
    common as csharp_common,
    verification as csharp_verification,
)
from aas_core_codegen import parse

import tests.common


class Test_against_recorded(unittest.TestCase):
    def test_cases(self) -> None:
        repo_dir = pathlib.Path(os.path.realpath(__file__)).parent.parent.parent.parent

        parent_case_dir = repo_dir / "dev/test_data/csharp/test_verification"
        assert parent_case_dir.exists() and parent_case_dir.is_dir(), parent_case_dir

        for model_pth in sorted(parent_case_dir.glob("**/meta_model.py")):
            text = model_pth.read_text(encoding="utf-8")
            atok, parse_exception = parse.source_to_atok(source=text)
            assert (
                parse_exception is None
            ), f"Unexpected parse exception in {model_pth}: {parse_exception=}"
            assert atok is not None

            lineno_columner = LinenoColumner(atok=atok)
            parsed_symbol_table, error = parse.atok_to_symbol_table(atok=atok)

            assert error is None, (
                f"Unexpected error in parsing {model_pth}: "
                f"{lineno_columner.error_message(error)}"
            )
            assert parsed_symbol_table is not None

            ir_symbol_table, error = intermediate.translate(
                parsed_symbol_table=parsed_symbol_table,
                atok=atok,
            )

            assert error is None, (
                f"Unexpected error in translating {model_pth} to intermediate: "
                f"{lineno_columner.error_message(error)}"
            )
            assert ir_symbol_table is not None

            code, errors = csharp_verification.generate(
                symbol_table=ir_symbol_table,
                namespace=csharp_common.NamespaceIdentifier("dummyNamespace"),
                spec_impls=dict(),
            )
            if errors is not None:
                joined_error = Error(
                    None, "Generating verification code failed", errors
                )
                raise AssertionError(
                    f"Unexpected errors in generating verification code "
                    f"for {model_pth}: {lineno_columner.error_message(joined_error)}"
                )
            assert code is not None

            expected_pth = model_pth.parent / "expected_verification.cs"
            if tests.common.RERECORD:
                expected_pth.write_text(code, encoding="utf-8")
            else:
                try:
                    expected_code = expected_pth.read_text(encoding="utf-8")
                except Exception as exception:
                    raise RuntimeError(
                        f"Failed to read the expected code from {expected_pth}"
                    ) from exception

                self.assertEqual(expected_code, code)


if __name__ == "__main__":
    unittest.main()



@abstract
class Another_grand_parent(DBC):
    pass


@abstract
class Grand_parent(DBC):
    pass


@abstract
class Parent(DBC, Grand_parent, Another_grand_parent):
    pass


@abstract
class Another_parent(DBC, Grand_parent, Another_grand_parent):
    pass


class Something(DBC, Parent, Another_parent):
    pass


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



"""
We encountered a bug when designing V3.0. When the class ``Lang_string`` is missing,
RDF generation breaks as we hard-wired certain blocks to generate specific code
for that particular class.

This unit test tests that RDF+SHACL generation does not break even though the class
``Lang_string`` is missing.
"""


@abstract
class Something_abstract(DBC):
    text: str

    def __init__(self, text: str) -> None:
        self.text = text


@invariant(
    lambda self: len(self.text) <= 128,
    "Text shall have a maximum length of 128 characters.",
)
class Something(Something_abstract, DBC):
    def __init__(self, text: str) -> None:
        Something_abstract.__init__(self, text=text)


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



class Something:
    some_str: str

    def __init__(
        self,
        some_str: str,
    ) -> None:
        self.some_str = some_str


__version__ = "V198.4"

__xml_namespace__ = "https://dummy/198/4"



"""Provide parsing functions to get started with testing some of the ReST rendering."""
import io
from typing import Optional

import docutils.nodes
import docutils.core


def parse_restructured_text(text: str) -> docutils.nodes.document:
    """Parse the given ReST ``text``."""
    warnings = io.StringIO()
    # noinspection PyUnusedLocal
    document = None  # type: Optional[docutils.nodes.document]

    document = docutils.core.publish_doctree(
        text, settings_overrides={"warning_stream": warnings}
    )

    warnings_text = warnings.getvalue()
    if warnings_text:
        raise RuntimeError(
            f"Failed to parse the description with docutils:\n"
            f"{warnings_text.strip()}\n\n"
            f"The original text was: {text!r}"
        )

    assert document is not None

    return document



class Item:
    pass


# fmt: off
@invariant(
    lambda self:
    len(self.some_property, 3) >= 1,
    "Some property must contain at least one item."
)
# fmt: on
class Something:
    some_property: List[Item]

    def __init__(self, some_property: List[Item]) -> None:
        self.some_property = some_property


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



#!/usr/bin/env python3

"""Check that the help snippets in the Readme coincide with the actual output."""
import argparse
import difflib
import os
import pathlib
import re
import subprocess
import sys
from typing import List, Tuple, Optional

import icontract


class Block:
    """Represent a block in the readme that needs to be checked."""

    @icontract.require(lambda command: command != "")
    @icontract.require(
        lambda start_line_idx, end_line_idx: start_line_idx <= end_line_idx
    )
    def __init__(self, command: str, start_line_idx: int, end_line_idx: int) -> None:
        """
        Initialize with the given values.

        :param command: help command
        :param start_line_idx: index of the first relevant line
        :param end_line_idx: index of the first line excluded from the block
        """
        self.command = command
        self.start_line_idx = start_line_idx
        self.end_line_idx = end_line_idx


HELP_STARTS_RE = re.compile(r"^.. Help starts: (?P<command>.*)$")


def parse_readme(lines: List[str]) -> Tuple[List[Block], List[str]]:
    """
    Parse the code blocks that represent help commands in the Readme.

    :param lines: lines of the readme file
    :return: (help blocks, errors if any)
    """
    blocks = []  # type: List[Block]
    errors = []  # type: List[str]

    i = 0
    while i < len(lines):
        mtch = HELP_STARTS_RE.match(lines[i])
        if mtch:
            command = mtch.group("command")
            help_ends = ".. Help ends: {}".format(command)
            try:
                end_index = lines.index(help_ends, i)
            except ValueError:
                end_index = -1

            if end_index == -1:
                return [], [
                    "Could not find_our_type the end marker {!r}".format(help_ends)
                ]

            blocks.append(
                Block(command=command, start_line_idx=i + 1, end_line_idx=end_index)
            )

            i = end_index + 1

        else:
            i += 1

    return blocks, errors


def capture_output_lines(command: str) -> List[str]:
    """Capture the output of a help command."""
    command_parts = command.split(" ")
    if command_parts[0] in ["python", "python3"]:
        # We need to replace "python" with "sys.executable" on Windows as
        # the environment is not properly inherited.
        command_parts[0] = sys.executable

    proc = None  # type: Optional[subprocess.Popen[str]]
    try:
        proc = subprocess.Popen(
            command_parts,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            encoding="utf-8",
        )
    except Exception as exception:
        raise RuntimeError(
            f"Failed to start the command {command_parts!r} "
            f"derived from the command {command!r} in the document"
        ) from exception

    assert proc is not None
    with proc:
        output, err = proc.communicate()
        if err:
            raise RuntimeError(
                f"The command {command!r} failed with exit code {proc.returncode} and "
                f"stderr:\n{err}"
            )

        return output.splitlines()


def output_lines_to_code_block(output_lines: List[str]) -> List[str]:
    """Translate the output of a help command to a RST code block."""
    result = (
        [".. code-block::", ""]
        + ["    " + output_line for output_line in output_lines]
        + [""]
    )

    result = [line.rstrip() for line in result]
    return result


def report_a_difference(
    got_lines: List[str], expected_lines: List[str]
) -> Optional[str]:
    """
    Report a difference between the ``got`` and ``expected``.

    Return None if no difference.
    """
    if got_lines == expected_lines:
        return None

    diff = difflib.ndiff(got_lines, expected_lines)
    return "\n".join(diff)


def main() -> int:
    """Execute the main routine."""
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "--overwrite",
        help="If set, overwrite the relevant part of README in-place.",
        action="store_true",
    )

    args = parser.parse_args()
    overwrite = bool(args.overwrite)

    repo_root = pathlib.Path(os.path.realpath(__file__)).parent.parent.parent
    pth = repo_root / "README.rst"

    text = pth.read_text(encoding="utf-8")
    lines = text.splitlines()

    blocks, errors = parse_readme(lines=lines)
    if errors:
        print("One or more errors in {}:".format(pth), file=sys.stderr)
        for error in errors:
            print(error, file=sys.stderr)

        return -1

    if len(blocks) == 0:
        return 0

    if overwrite:
        result = []  # type: List[str]

        previous_block = None  # type: Optional[Block]
        for block in blocks:
            output_lines = capture_output_lines(command=block.command)
            code_block_lines = output_lines_to_code_block(output_lines=output_lines)

            if previous_block is None:
                result.extend(lines[: block.start_line_idx])
            else:
                result.extend(lines[previous_block.end_line_idx : block.start_line_idx])

            result.extend(code_block_lines)
            previous_block = block

        assert previous_block is not None
        result.extend(lines[previous_block.end_line_idx :])
        result.append("")  # new line at the end of file

        pth.write_text("\n".join(result), encoding="utf-8")

    else:
        for block in blocks:
            output_lines = capture_output_lines(command=block.command)
            code_block_lines = output_lines_to_code_block(output_lines=output_lines)

            expected_lines = lines[block.start_line_idx : block.end_line_idx]
            expected_lines = [line.rstrip() for line in expected_lines]

            diff_error = report_a_difference(
                got_lines=code_block_lines, expected_lines=expected_lines
            )
            if diff_error is not None:
                print(diff_error, file=sys.stderr)
                return -1

    return 0


if __name__ == "__main__":
    sys.exit(main())



class Something:
    @snapshot(lambda x, y: x + y > 0, name="0932invalid")
    def do_something(self, x: int, y: int) -> int:
        pass


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



@abstract
@serialization(with_model_type=True)
class Something_abstract:
    some_str: str

    def __init__(self, some_str: str) -> None:
        self.some_str = some_str


class Something_concrete(Something_abstract):
    something_str: str

    def __init__(self, some_str: str, something_str: str) -> None:
        Something_abstract.__init__(self, some_str)

        self.something_str = something_str


class Another_concrete(Something_abstract):
    another_str: str

    def __init__(self, some_str: str, another_str: str) -> None:
        Something_abstract.__init__(self, some_str)

        self.another_str = another_str


class Container:
    something_abstract: Something_abstract

    def __init__(self, something_abstract: Something_abstract) -> None:
        self.something_abstract = something_abstract


__version__ = "V198.4"

__xml_namespace__ = "https://dummy/198/4"



"""Download the latest meta-model for V3.0 to the test data."""

import argparse
import os
import pathlib
import sys
from typing import Final, Mapping, Union

import black
import requests

OWNER: Final[str] = "aas-core-works"
REPO: Final[str] = "aas-core-meta"
REF: Final[str] = "main"
REMOTE_PATH: Final[str] = "aas_core_meta/v3.py"


def _latest_commit_sha(timeout: float = 15.0) -> str:
    """Resolve the latest commit SHA on the ``REF`` for a specific path."""
    params: Mapping[str, Union[str, int]] = {
        "path": REMOTE_PATH,
        "sha": REF,
        "per_page": 1,
    }
    resp = requests.get(
        f"https://api.github.com/repos/{OWNER}/{REPO}/commits",
        params=params,
        timeout=timeout,
    )
    try:
        resp.raise_for_status()
    except requests.HTTPError as ex:
        raise RuntimeError(
            f"Failed to resolve latest commit for "
            f"{OWNER}/{REPO}:{REF} path={REMOTE_PATH} "
            f"({resp.status_code}): {resp.text}"
        ) from ex

    commits = resp.json()
    if not commits or not isinstance(commits, list):
        raise RuntimeError(
            "Could not determine latest commit SHA (empty API response)."
        )

    sha = commits[0].get("sha")
    if not isinstance(sha, str):
        raise RuntimeError("API did not return a valid commit SHA.")

    if len(sha) == 0:
        raise RuntimeError("API returned an empty commit SHA.")

    return sha


def _raw_url(sha: str) -> str:
    """Get the URL of the raw file on GitHub."""
    return f"https://raw.githubusercontent.com/{OWNER}/{REPO}/{sha}/{REMOTE_PATH}"


def _download(url: str, timeout: float = 30.0) -> str:
    """Download the raw file contents."""
    resp = requests.get(url, timeout=timeout)
    try:
        resp.raise_for_status()
    except requests.HTTPError as ex:
        raise RuntimeError(
            f"Failed to download file from {url} ({resp.status_code}): {resp.text}"
        ) from ex

    return resp.text


def retrieve_sha_and_download() -> str:
    """
    Get the latest SHA revision of the meta-model and download it to test data.

    Return the SHA of the downloaded meta-model.
    """
    sha = _latest_commit_sha()

    raw_url = _raw_url(sha)

    content = _download(url=raw_url)

    banner = f"# Downloaded from: {raw_url}\n# Do NOT edit or append!"

    repo_root = pathlib.Path(os.path.realpath(__file__)).parent.parent.parent

    out_path = repo_root / "dev/test_data/real_meta_models/aas_core_meta.v3.py"
    out_path.parent.mkdir(exist_ok=True)

    out_path.write_text(f"{banner}\n\n{content.rstrip()}\n\n{banner}", encoding="utf-8")

    black.format_file_in_place(
        src=out_path, fast=False, mode=black.FileMode(), write_back=black.WriteBack.YES
    )

    print(f"Wrote and reformatted to: {out_path} (from commit {sha[:8]}).")

    return sha


def main() -> int:
    """Execute the main routine."""
    parser = argparse.ArgumentParser(description=__doc__)
    _ = parser.parse_args()

    retrieve_sha_and_download()

    return 0


if __name__ == "__main__":
    sys.exit(main())



class Something:
    """Represent something."""

    some_property: int
    """some property"""

    another_property: str

    yet_another_property: str
    """yet another property"""


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



# We test for ``\x09`` and other special characters, and make sure they are represented
# adequately in the string dump of the symbol table.


class Non_empty_XML_serializable_string(str, DBC):
    r"""
    Represent a string with at least one character.

    The string should also be serializable to XML, which is the background for
    the following constraint.

    :constraint AASd-130:

        An attribute with data type "string" shall consist of these characters only:
        ^[\x09\x0A\x0D\x20-\uD7FF\uE000-\uFFFD\u00010000-\u0010FFFF]*$.
    """


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



Something: Set[str] = constant_set(values=["some literal", "another literal"])

Something_extended: Set[str] = constant_set(
    values=[
        "some literal",
        "another literal",
        "yet another literal",
    ],
    superset_of=[Something],
)


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



@abstract
class VeryAbstract:
    some_property: int

    @require(lambda some_property: some_property > 0)
    def __init__(self, some_property: int) -> None:
        self.some_property = some_property

    def some_func(self) -> None:
        """Do something."""


@abstract
class SomethingAbstract(VeryAbstract):
    another_property: int

    @require(lambda another_property: another_property > 0)
    def __init__(self, some_property: int, another_property: int) -> None:
        VeryAbstract.__init__(self, some_property)
        self.another_property = another_property

    def another_func(self) -> None:
        """Do something else."""


class Concrete(SomethingAbstract):
    yet_another_property: int

    @require(lambda yet_another_property: yet_another_property > 0)
    def __init__(
        self, some_property: int, another_property: int, yet_another_property: int
    ) -> None:
        SomethingAbstract.__init__(self, some_property, another_property)
        self.yet_another_property = yet_another_property

    def yet_another_func(self) -> None:
        """Do yet something else."""


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



# Automatically generated with python_protobuf/main.py.
# Do NOT edit or append.


"""Convert instances from and to Protocol Buffers."""


from typing import Mapping, TypeVar

import google.protobuf.message

from aas_core3 import types
import aas_core3_protobuf.types_pb2 as types_pb


# region From Protocol Buffers


def something_from_pb(
    that: types_pb.Something
) -> types.Something:
    """
    Parse ``that`` Protocol Buffer to an instance of a concrete class.

    Example usage:

    .. code-block::

        import aas_core3_protobuf.types_pb2 as types_pb
        from aas_core3_protobuf.pbization import something_from_pb

        some_bytes = b'... serialized types_pb.Something ...'
        something_pb = types_pb.Something()
        something_pb.FromString(
            some_bytes
        )

        something = something_from_pb(
            something_pb
        )
        # Do something with the something...

    """
    return types.Something(
        some_bool=that.some_bool,
        some_int=that.some_int,
        some_float=that.some_float,
        some_str=that.some_str,
        some_byte_array=bytearray(that.some_byte_array),
        some_optional_bool=(
            that.some_optional_bool
            if that.HasField('some_optional_bool')
            else None
        ),
        some_optional_int=(
            that.some_optional_int
            if that.HasField('some_optional_int')
            else None
        ),
        some_optional_float=(
            that.some_optional_float
            if that.HasField('some_optional_float')
            else None
        ),
        some_optional_str=(
            that.some_optional_str
            if that.HasField('some_optional_str')
            else None
        ),
        some_optional_byte_array=(
            bytearray(that.some_optional_byte_array)
            if that.HasField('some_optional_byte_array')
            else None
        )
    )


# fmt: off
_FROM_PB_MAP = {
    types_pb.Something:
        something_from_pb
}
# fmt: on


def from_pb(
    that: google.protobuf.message.Message
) -> types.Class:
    """
    Parse ``that`` Protocol Buffer into a model instance.

    The concrete parsing is determined based on the runtime type of ``that``
    Protocol Buffer. It is assumed that ``that`` is an instance of a message
    coming from the Protocol Buffer definitions corresponding to the meta-model.

    Example usage:

    .. code-block::

        import aas_core3_protobuf.types_pb2 as types_pb
        from aas_core3_protobuf.pbization import from_pb

        some_bytes = b'... serialized types_pb.Something ...'
        instance_pb = types_pb.Something()
        instance_pb.FromString(
            some_bytes
        )

        instance = from_pb(
            instance_pb
        )
        # Do something with the instance...
    """
    from_pb_func = _FROM_PB_MAP.get(that.__class__, None)

    if from_pb_func is None:
        raise ValueError(
            f"We do not know how to parse the protocol buffer "
            f"of type {that.__class__} into a model instance."
        )

    result = from_pb_func(that)  # type: ignore
    assert isinstance(result, types.Class)
    return result


# endregion From Protocol Buffers


# region To Protocol Buffers


T = TypeVar("T")


class _PartialVisitorWithContext(types.AbstractVisitorWithContext[T]):
    """
    Visit instances in context with double-dispatch.

    This class is meant to be inherited from. If you do not override a method,
    it will raise an exception. This is a partial visitor, meaning that some
    visits are unexpected by design.
    """
    # pylint: disable=missing-docstring

    def visit_something_with_context(
        self,
        that: types.Something,
        context: T
    ) -> None:
        raise AssertionError(f"Unexpected visitation of {that.__class__}")


def something_to_pb(
    that: types.Something,
    target: types_pb.Something
) -> None:
    """
    Set fields in ``target`` based on ``that`` instance.

    Example usage:

    .. code-block::

        import aas_core3.types as types

        import aas_core3_protobuf.types_pb2 as types_pb
        from aas_core3_protobuf.pbization import something_to_pb

        something = types.Something(
            ... # some constructor arguments
        )

        something_pb = types_pb.Something()
        something_to_pb(
            something,
            something_pb
        )

        some_bytes = something_pb.SerializeToString()
        # Do something with some_bytes. For example, transfer them
        # over the wire.
    """
    target.some_bool = that.some_bool

    target.some_int = that.some_int

    target.some_float = that.some_float

    target.some_str = that.some_str

    target.some_byte_array = bytes(that.some_byte_array)

    if that.some_optional_bool is not None:
        target.some_optional_bool = that.some_optional_bool

    if that.some_optional_int is not None:
        target.some_optional_int = that.some_optional_int

    if that.some_optional_float is not None:
        target.some_optional_float = that.some_optional_float

    if that.some_optional_str is not None:
        target.some_optional_str = that.some_optional_str

    if that.some_optional_byte_array is not None:
        target.some_optional_byte_array = bytes(that.some_optional_byte_array)


class _ToPbTransformer(
    types.AbstractTransformer[google.protobuf.message.Message]
):
    """
    Dispatch to-pb conversion to the concrete functions.

    The classes with descendants (i.e., subtypes) are always going to be converted
    to their concrete Protocol Buffer instead of the choice (union) Protocol Buffer
    class. We made this decision with the compactness of messages in mind.
    """
    def transform_something(
        self,
        that: types.Something
    ) -> google.protobuf.message.Message:
        """
        Convert ``that`` instance
        to a :py:class:`types_pb.Something`.
        """
        result = types_pb.Something()
        something_to_pb(that, result)
        return result


_TO_PB_TRANSFORMER = _ToPbTransformer()


def to_pb(
    that: types.Class,
) -> google.protobuf.message.Message:
    """
    Dispatch to-pb conversion to the concrete functions.

    The classes with descendants (i.e., subtypes) are always going to be converted
    to their concrete Protocol Buffer message type instead of the choice (union) type.
    We made this decision with the compactness of messages in mind as choice types
    would occupy a tiny bit more space.

    Example usage:

    .. code-block::

        import aas_core3.types as types

        from aas_core3_protobuf.pbization import to_pb

        instance = types.Something(
            ... # some constructor arguments
        )

        instance_pb = to_pb(
            instance
        )

        some_bytes = instance_pb.SerializeToString()
        # Do something with some_bytes. For example, transfer them
        # over the wire.
    """
    return _TO_PB_TRANSFORMER.transform(that)


# endregion To Protocol Buffers


# Automatically generated with python_protobuf/main.py.
# Do NOT edit or append.



# pylint: disable=missing-docstring

import textwrap
import unittest
from typing import List

import tests.common
from aas_core_codegen import intermediate
from aas_core_codegen.common import Identifier
from aas_core_codegen.intermediate import type_inference as intermediate_type_inference


class Test_with_smoke(unittest.TestCase):
    @staticmethod
    def execute(source: str) -> None:
        """Execute a smoke test on all the invariants of all the classes."""
        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)

        assert symbol_table is not None

        base_environment = intermediate_type_inference.populate_base_environment(
            symbol_table=symbol_table
        )

        for our_type in symbol_table.our_types:
            if isinstance(
                our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
            ):
                environment = intermediate_type_inference.MutableEnvironment(
                    parent=base_environment
                )
                environment.set(
                    Identifier("self"),
                    intermediate_type_inference.OurTypeAnnotation(our_type=our_type),
                )

                for invariant in our_type.invariants:
                    # fmt: off
                    _, inference_error = (
                        intermediate_type_inference.infer_for_invariant(
                            invariant=invariant,
                            environment=environment
                        )
                    )
                    # fmt: on

                    assert (
                        inference_error is None
                    ), tests.common.most_underlying_messages([inference_error])

        for verification in symbol_table.verification_functions:
            if not isinstance(verification, intermediate.TranspilableVerification):
                continue

            # fmt: off
            _, inference_error = (
                intermediate_type_inference.infer_for_verification(
                    verification=verification,
                    base_environment=base_environment
                )
            )
            # fmt: on

            assert inference_error is None, tests.common.most_underlying_messages(
                [inference_error]
            )

    def expect_type_inference_to_fail(
        self, source: str, expected_joined_message: str
    ) -> None:
        """Execute a smoke test and expect type inference to fail."""
        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)

        assert symbol_table is not None

        base_environment = intermediate_type_inference.populate_base_environment(
            symbol_table=symbol_table
        )

        type_inference_errors = []  # type: List[str]

        for our_type in symbol_table.our_types:
            if isinstance(
                our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
            ):
                environment = intermediate_type_inference.MutableEnvironment(
                    parent=base_environment
                )
                environment.set(
                    Identifier("self"),
                    intermediate_type_inference.OurTypeAnnotation(our_type=our_type),
                )

                for invariant in our_type.invariants:
                    # fmt: off
                    _, inference_error = (
                        intermediate_type_inference.infer_for_invariant(
                            invariant=invariant,
                            environment=environment
                        )
                    )
                    # fmt: on

                    if inference_error is not None:
                        type_inference_errors.append(
                            tests.common.most_underlying_messages([inference_error])
                        )

        for verification in symbol_table.verification_functions:
            if not isinstance(verification, intermediate.TranspilableVerification):
                continue

            # fmt: off
            _, inference_error = (
                intermediate_type_inference.infer_for_verification(
                    verification=verification,
                    base_environment=base_environment
                )
            )
            # fmt: on

            if inference_error is not None:
                type_inference_errors.append(
                    tests.common.most_underlying_messages([inference_error])
                )

        assert len(type_inference_errors) > 0, (
            f"Expected one or more type inference errors, "
            f"but got none on the source code:\n{source}"
        )

        joined_message = "\n".join(type_inference_errors)
        self.assertEqual(expected_joined_message, joined_message, source)

    def test_enumeration_literal_as_member(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_enum(Enum):
                Literal_a = "LITERAL-A"
                Literal_b = "LITERAL-B"
                Literal_c = "LITERAL-C"

            @invariant(
                lambda self:
                self.something == Some_enum.Literal_a
                or self.something == Some_enum.Literal_b,
                "Something must be either LITERAL-A or LITERAL-B."
            )
            class Some_class:
                something: Some_enum

                def __init__(self, something: Some_enum) -> None:
                    self.something = something


            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        Test_with_smoke.execute(source=source)

    def test_class_member(self) -> None:
        source = textwrap.dedent(
            """\
            @invariant(
                lambda self:
                self.something >= 1,
                "Something must be at least 1."
            )
            class Some_class:
                something: int

                def __init__(self, something: int) -> None:
                    self.something = something


            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        Test_with_smoke.execute(source=source)

    def test_non_nullness_of_members_member_in_implication(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_class:
                something: Optional[str]

                def __init__(self, something: Optional[str] = None) -> None:
                    self.something = something

            @invariant(
                lambda self:
                not (
                    self.some_instance is not None
                    and self.some_instance.something is not None
                ) or (
                    self.some_instance.something == "some-literal"
                ),
                "If something of some instance is defined, it must be set "
                "to some-literal."
            )
            class Another_class:
                some_instance: Optional[Some_class]

                def __init__(self, some_instance: Optional[Some_class] = None) -> None:
                    self.some_instance = some_instance


            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        Test_with_smoke.execute(source=source)

    def test_non_nullness_of_members_member_in_conjunction(self) -> None:
        source = textwrap.dedent(
            """\
            class Some_class:
                something: Optional[str]

                def __init__(self, something: Optional[str] = None) -> None:
                    self.something = something

            @invariant(
                lambda self:
                self.some_instance is not None
                and self.some_instance.something is not None
                and self.some_instance.something == "some-literal",
                "Something of some instance must be defined and set to some-literal."
            )
            class Another_class:
                some_instance: Optional[Some_class]

                def __init__(self, some_instance: Optional[Some_class] = None) -> None:
                    self.some_instance = some_instance


            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        Test_with_smoke.execute(source=source)

    def test_non_nullness_in_disjunction_with_is_none(self) -> None:
        source = textwrap.dedent(
            """\
            @invariant(
                lambda self:
                (self.some_property is None) or (self.some_property == "dummy"),
                "Dummy description"
            )
            class Something:
                some_property: Optional[str]

                def __init__(self, some_property: Optional[str] = None) -> None:
                    self.some_property = some_property


            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        Test_with_smoke.execute(source=source)

    def test_is_none_fails_on_non_optional(self) -> None:
        source = textwrap.dedent(
            """\
            @invariant(
                lambda self:
                self.something is None,
                "Dummy invariant description"
            )
            class Some_class:
                something: str

                def __init__(self, something: str) -> None:
                    self.something = something

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        self.expect_type_inference_to_fail(
            source=source,
            expected_joined_message=(
                "Expected the value to be of an optional type "
                "for a nullness check (``is None``), but got str"
            ),
        )

    def test_is_not_none_fails_on_non_optional(self) -> None:
        source = textwrap.dedent(
            """\
            @invariant(
                lambda self:
                self.something is not None,
                "Dummy invariant description"
            )
            class Some_class:
                something: str

                def __init__(self, something: str) -> None:
                    self.something = something

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            """
        )

        self.expect_type_inference_to_fail(
            source=source,
            expected_joined_message=(
                "Expected the value to be of an optional type "
                "for a non-nullness check (``is not None``), but got str"
            ),
        )


if __name__ == "__main__":
    unittest.main()



class Concrete:
    x: int

    @require(lambda x: x > 0)
    def __init__(self, x: int) -> None:
        self.x = x

    @require(lambda self: self.x > 2)
    @require(lambda number: number > 0)
    def some_func(self, number: int) -> int:
        """Do something."""


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



# pylint: disable=missing-docstring

import unittest

import aas_core_codegen.python.common as python_common


class TestStringLiteral(unittest.TestCase):
    def test_empty(self) -> None:
        self.assertEqual("''", python_common.string_literal(""))

    def test_no_quotes(self) -> None:
        self.assertEqual("'x'", python_common.string_literal("x"))

    def test_double_quotes_but_no_single_quotes(self) -> None:
        self.assertEqual("'a\"b'", python_common.string_literal('a"b'))

    def test_single_quotes_but_no_double_quotes(self) -> None:
        self.assertEqual('"a\'b"', python_common.string_literal("a'b"))

    def test_equal_number_of_single_quotes_and_double_quotes(self) -> None:
        self.assertEqual("'a\\'\"b'", python_common.string_literal("a'\"b"))

    def test_more_single_quotes_than_double_quotes(self) -> None:
        self.assertEqual('"a\'\'\\"b"', python_common.string_literal("a''\"b"))

    def test_less_single_quotes_than_double_quotes(self) -> None:
        self.assertEqual("'a\\'\"\"b'", python_common.string_literal('a\'""b'))

    def test_tab(self) -> None:
        self.assertEqual("'a\\tb'", python_common.string_literal("a\tb"))

    def test_unset_duplicate_curly_brackets_no_quotes(self) -> None:
        self.assertEqual(
            "'{AAA}'",
            python_common.string_literal("{AAA}", duplicate_curly_brackets=False),
        )

        self.assertEqual("'{AAA}'", python_common.string_literal("{AAA}"))

    def test_duplicate_curly_brackets_no_quotes(self) -> None:
        self.assertEqual(
            "'{{AAA}}'",
            python_common.string_literal("{AAA}", duplicate_curly_brackets=True),
        )

    def test_unset_duplicate_curly_brackets_single_quote(self) -> None:
        self.assertEqual(
            '"{A\'A}"',
            python_common.string_literal("{A'A}", duplicate_curly_brackets=False),
        )

        self.assertEqual('"{A\'A}"', python_common.string_literal("{A'A}"))

    def test_duplicate_curly_brackets_single_quote(self) -> None:
        self.assertEqual(
            '"{{A\'A}}"',
            python_common.string_literal("{A'A}", duplicate_curly_brackets=True),
        )

    def test_unset_duplicate_curly_brackets_double_quote(self) -> None:
        self.assertEqual(
            "'{A\"A}'",
            python_common.string_literal('{A"A}', duplicate_curly_brackets=False),
        )

        self.assertEqual("'{A\"A}'", python_common.string_literal('{A"A}'))

    def test_duplicate_curly_brackets_double_quote(self) -> None:
        self.assertEqual(
            "'{{A\"A}}'",
            python_common.string_literal('{A"A}', duplicate_curly_brackets=True),
        )


class TestBytesLiteral(unittest.TestCase):
    def test_empty(self) -> None:
        self.assertTupleEqual(
            ('b""', False), python_common.bytes_literal(bytearray(b""))
        )

    def test_one(self) -> None:
        self.assertTupleEqual(
            ('b"\\x00"', False), python_common.bytes_literal(bytearray(b"\x00"))
        )

    def test_eight(self) -> None:
        self.assertTupleEqual(
            ('b"\\x00\\x01\\x02\\x03\\x04\\x05\\x06\\x07"', False),
            python_common.bytes_literal(bytearray(b"\x00\x01\x02\x03\x04\x05\x06\x07")),
        )

    def test_nine(self) -> None:
        self.assertTupleEqual(
            (
                '''\
b"\\x00\\x01\\x02\\x03\\x04\\x05\\x06\\x07"
b"\\x08"''',
                True,
            ),
            python_common.bytes_literal(
                bytearray(b"\x00\x01\x02\x03\x04\x05\x06\x07\x08")
            ),
        )

    def test_sixteen(self) -> None:
        self.assertTupleEqual(
            (
                '''\
b"\\x00\\x01\\x02\\x03\\x04\\x05\\x06\\x07"
b"\\x08\\x09\\x10\\x11\\x12\\x13\\x14\\x15"''',
                True,
            ),
            python_common.bytes_literal(
                bytearray(
                    b"\x00\x01\x02\x03\x04\x05\x06\x07"
                    b"\x08\x09\x10\x11\x12\x13\x14\x15"
                )
            ),
        )


if __name__ == "__main__":
    unittest.main()



@abstract
class Something_abstract:
    some_str: str

    def __init__(self, some_str: str) -> None:
        self.some_str = some_str


@abstract
class Another_abstract:
    another_str: str

    def __init__(self, another_str: str) -> None:
        self.another_str = another_str


class Concrete(Something_abstract, Another_abstract):
    def __init__(self, some_str: str, another_str: str) -> None:
        Something_abstract.__init__(self, some_str)
        Another_abstract.__init__(self, another_str)


__version__ = "V198.4"

__xml_namespace__ = "https://dummy/198/4"



# pylint: disable=missing-docstring

import textwrap
import unittest.mock

# noinspection PyProtectedMember
import aas_core_codegen.csharp.description as csharp_description
import tests.common
from aas_core_codegen.common import Stripped, Identifier


class Test_to_render_description_of_meta_model(unittest.TestCase):
    @staticmethod
    def render(source: str) -> Stripped:
        """Generate the C# description comment based on ``source``."""
        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)
        assert symbol_table is not None

        assert (
            symbol_table.meta_model.description is not None
        ), "Expected a meta-model description, but found none."

        code, errors = csharp_description.generate_comment_for_meta_model(
            symbol_table.meta_model.description
        )
        assert errors is None, tests.common.most_underlying_messages(errors)

        assert code is not None

        return code

    def test_empty_description_not_allowed(self) -> None:
        source = textwrap.dedent(
            '''\
            """"""  # Intentionally left empty

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            '''
        )

        _, error = tests.common.translate_source_to_intermediate(source=source)
        assert error is not None
        self.assertEqual(
            "Unexpected empty description", tests.common.most_underlying_messages(error)
        )

    def test_only_summary(self) -> None:
        comment_code = self.__class__.render(
            textwrap.dedent(
                '''\
                """Do & drink something."""

                __version__ = "dummy"
                __xml_namespace__ = "https://dummy.com"
                '''
            )
        )

        self.assertEqual(
            textwrap.dedent(
                """\
                /// <summary>
                /// Do &amp; drink something.
                /// </summary>"""
            ),
            comment_code,
        )


class Test_to_render_description_of_our_types(unittest.TestCase):
    @staticmethod
    def render(source: str) -> Stripped:
        """
        Generate the C# description comment based on ``source``.

        The ``source`` is expected to contain only a single class, a class
        ``Some_class``.
        """
        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)
        assert symbol_table is not None

        something_type = symbol_table.must_find_our_type(Identifier("Something"))
        assert something_type.description is not None

        code, errors = csharp_description.generate_comment_for_our_type(
            something_type.description
        )
        assert errors is None, tests.common.most_underlying_messages(errors)

        assert code is not None

        return code

    def test_empty_description_not_allowed(self) -> None:
        source = textwrap.dedent(
            '''\
            class Some_class:
                """"""  # Intentionally left empty

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            '''
        )

        _, error = tests.common.translate_source_to_intermediate(source=source)
        assert error is not None
        self.assertEqual(
            "Unexpected empty description", tests.common.most_underlying_messages(error)
        )

    def test_no_summary(self) -> None:
        source = textwrap.dedent(
            '''\
            class Some_class:
                """
                * Some
                * Bullet
                * List
                """

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            '''
        )

        _, error = tests.common.translate_source_to_intermediate(source=source)

        assert error is not None

        self.assertEqual(
            "Expected the first document element to be a summary and thus a paragraph, "
            "but got: "
            '<bullet_list bullet="*">'
            "<list_item><paragraph>Some</paragraph></list_item>"
            "<list_item><paragraph>Bullet</paragraph></list_item>"
            "<list_item><paragraph>List</paragraph></list_item>"
            "</bullet_list>",
            tests.common.most_underlying_messages(error),
        )

    def test_unexpected_directive(self) -> None:
        source = textwrap.dedent(
            '''\
            class Some_class:
                """
                Do something.

                :someDirective:
                    I am unexpected.
                """

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            '''
        )

        _, error = tests.common.translate_source_to_intermediate(source=source)

        assert error is not None

        self.assertEqual(
            "Expected only directives such as ``constraint some-identifier`` "
            "in this context, but got a directive with 1 part(s): someDirective",
            tests.common.most_underlying_messages(error),
        )

    def test_only_summary(self) -> None:
        comment_code = Test_to_render_description_of_our_types.render(
            textwrap.dedent(
                '''\
                class Something:
                    """Do & drink something."""

                __version__ = "dummy"
                __xml_namespace__ = "https://dummy.com"
                '''
            )
        )

        self.assertEqual(
            textwrap.dedent(
                """\
                /// <summary>
                /// Do &amp; drink something.
                /// </summary>"""
            ),
            comment_code,
        )

    def test_summary_with_class_reference(self) -> None:
        comment_code = Test_to_render_description_of_our_types.render(
            textwrap.dedent(
                '''\
                class Something:
                    """Do & drink :class:`Something`."""

                __version__ = "dummy"
                __xml_namespace__ = "https://dummy.com"
                '''
            )
        )

        self.assertEqual(
            """\
/// <summary>
/// Do &amp; drink <see cref="Aas.Something" />.
/// </summary>""",
            comment_code,
        )

    def test_summary_with_interface_reference(self) -> None:
        comment_code = Test_to_render_description_of_our_types.render(
            textwrap.dedent(
                '''\
                @abstract
                class Something:
                    """Do & drink :class:`Something`."""

                __version__ = "dummy"
                __xml_namespace__ = "https://dummy.com"
                '''
            )
        )

        self.assertEqual(
            textwrap.dedent(
                """\
                /// <summary>
                /// Do &amp; drink <see cref="Aas.ISomething" />.
                /// </summary>"""
            ),
            comment_code,
        )

    def test_summary_with_enumeration_reference(self) -> None:
        comment_code = Test_to_render_description_of_our_types.render(
            textwrap.dedent(
                '''\
                class Something(Enum):
                    """Do & drink :class:`Something`."""

                __version__ = "dummy"
                __xml_namespace__ = "https://dummy.com"
                '''
            )
        )

        self.assertEqual(
            textwrap.dedent(
                """\
                /// <summary>
                /// Do &amp; drink <see cref="Aas.Something" />.
                /// </summary>"""
            ),
            comment_code,
        )

    def test_summary_and_remarks(self) -> None:
        comment_code = Test_to_render_description_of_our_types.render(
            textwrap.dedent(
                '''\
                class Something:
                    """
                    Do & drink something.

                    First & remark.

                    Second & remark.
                    """

                __version__ = "dummy"
                __xml_namespace__ = "https://dummy.com"
                '''
            )
        )

        self.assertEqual(
            """\
/// <summary>
/// Do &amp; drink something.
/// </summary>
/// <remarks>
/// <para>
/// First &amp; remark.
/// </para>
/// <para>
/// Second &amp; remark.
/// </para>
/// </remarks>""",
            comment_code,
        )

    def test_summary_remarks_and_constraints(self) -> None:
        comment_code = Test_to_render_description_of_our_types.render(
            # NOTE (mristin, 2022-07-21):
            # We explicitly test here for three cases:
            # 1) a single-paragraph constraint,
            # 2) a two-paragraph constraint, and
            # 3) a constraint with an unordered list.
            textwrap.dedent(
                '''\
                class Something:
                    """
                    Do & drink something.

                    First & remark.

                    :constraint AAS-001:
                        You shall parse.

                    :constraint AAS-002:
                        You have to do something.

                        Really.

                    :constraint AAS-003:
                        You shall:

                        * Do something,
                        * And do it now.
                    """

                __version__ = "dummy"
                __xml_namespace__ = "https://dummy.com"
                '''
            )
        )
        self.assertEqual(
            """\
/// <summary>
/// Do &amp; drink something.
/// </summary>
/// <remarks>
/// <para>
/// First &amp; remark.
/// </para>
/// <para>
/// Constraints:
/// </para>
/// <ul>
///   <li>
///     Constraint AAS-001:
///     You shall parse.
///   </li>
///   <li>
///     <para>
///     Constraint AAS-002:
///     You have to do something.
///     </para>
///     <para>
///     Really.
///     </para>
///   </li>
///   <li>
///     <para>
///     Constraint AAS-003:
///     You shall:
///     </para>
///     <ul>
///       <li>
///         Do something,
///       </li>
///       <li>
///         And do it now.
///       </li>
///     </ul>
///   </li>
/// </ul>
/// </remarks>""",
            comment_code,
        )

    def test_summary_and_note_in_the_remarks(self) -> None:
        # NOTE (mristin, 2022-07-21):
        # This is a real-world example from aas-core-meta which is kept here as
        # a regression test as we did not render it correctly at first. There was
        # a ``<para>`` nested in the ``<remarks>`` element.
        comment_code = Test_to_render_description_of_our_types.render(
            textwrap.dedent(
                '''\
                class Something:
                    """
                    Global reference to the data specification template used by
                    the element.

                    .. note::

                        This is a global reference.
                    """

                __version__ = "dummy"
                __xml_namespace__ = "https://dummy.com"
                '''
            )
        )

        self.assertEqual(
            """\
/// <summary>
/// Global reference to the data specification template used by
/// the element.
/// </summary>
/// <remarks>
/// This is a global reference.
/// </remarks>""",
            comment_code,
        )


class Test_to_render_description_of_signature(unittest.TestCase):
    @staticmethod
    def render(source: str) -> Stripped:
        """
        Generate the C# description comment based on ``source``.

        The ``source`` is expected to contain only a single verification function,
        ``verify_something``.
        """
        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)
        assert symbol_table is not None

        verification = symbol_table.verification_functions_by_name[
            Identifier("verify_something")
        ]

        assert verification.description is not None

        code, errors = csharp_description.generate_comment_for_signature(
            verification.description
        )
        assert errors is None, tests.common.most_underlying_messages(errors)

        assert code is not None

        return code

    def test_empty_description_not_allowed(self) -> None:
        source = textwrap.dedent(
            '''\
            @verification
            def verify_something(text: str) -> bool:
                """"""  # Intentionally left empty
                return match(r'.*', text) is not None

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            '''
        )

        _, error = tests.common.translate_source_to_intermediate(source=source)

        assert error is not None

        self.assertEqual(
            "Unexpected empty description", tests.common.most_underlying_messages(error)
        )

    def test_only_summary(self) -> None:
        source = textwrap.dedent(
            '''\
            @verification
            def verify_something(text: str) -> bool:
                """Verify something."""
                return match(r'^.*$', text) is not None

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            '''
        )

        code = Test_to_render_description_of_signature.render(source=source)

        self.assertEqual(
            textwrap.dedent(
                """\
                /// <summary>
                /// Verify something.
                /// </summary>"""
            ),
            code,
        )

    def test_params_and_returns(self) -> None:
        # NOTE (mristin, 2022-07-21):
        # We explicitly check here for multiple paragraphs in the param and returns.
        source = textwrap.dedent(
            '''\
            @verification
            def verify_something(first: str, second: str) -> bool:
                """
                Verify something.

                :param first: to be checked
                :param second:
                    another thing to be checked.

                    really to be checked.

                :returns:
                    True if :paramref:`first` and :paramref:`second` are
                    a valid something.

                    Otherwise, return false.
                """
                return match(r'.*', text) is not None

            __version__ = "dummy"
            __xml_namespace__ = "https://dummy.com"
            '''
        )

        code = Test_to_render_description_of_signature.render(source=source)

        self.assertEqual(
            """\
/// <summary>
/// Verify something.
/// </summary>
/// <param name="first">
/// to be checked
/// </param>
/// <param name="second">
/// <para>
/// another thing to be checked.
/// </para>
/// <para>
/// really to be checked.
/// </para>
/// </param>
/// <returns>
/// <para>
/// True if <paramref name="first" /> and <paramref name="second" /> are
/// a valid something.
/// </para>
/// <para>
/// Otherwise, return false.
/// </para>
/// </returns>""",
            code,
        )


class Test_to_render_paragraphs(unittest.TestCase):
    @staticmethod
    def render(source: str) -> Stripped:
        """Generate the C# description comment based on ``source``."""
        symbol_table, error = tests.common.translate_source_to_intermediate(
            source=source
        )
        assert error is None, tests.common.most_underlying_messages(error)
        assert symbol_table is not None

        assert (
            symbol_table.meta_model.description is not None
        ), "Expected a meta-model description, but found none."

        code, errors = csharp_description.generate_comment_for_meta_model(
            symbol_table.meta_model.description
        )
        assert errors is None, tests.common.most_underlying_messages(errors)

        assert code is not None

        return code

    def test_single_paragraph(self) -> None:
        comment_code = self.__class__.render(
            textwrap.dedent(
                '''\
                """Write a single paragraph."""

                __version__ = "dummy"
                __xml_namespace__ = "https://dummy.com"
                '''
            )
        )

        self.assertEqual(
            textwrap.dedent(
                """\
                /// <summary>
                /// Write a single paragraph.
                /// </summary>"""
            ),
            comment_code,
        )

    def test_multiple_text_remarks(self) -> None:
        comment_code = self.__class__.render(
            textwrap.dedent(
                '''\
                """
                This is summary.

                This is first paragraph of remarks.

                This is second paragraph of remarks.
                """

                __version__ = "dummy"
                __xml_namespace__ = "https://dummy.com"
                '''
            )
        )

        self.assertEqual(
            """\
/// <summary>
/// This is summary.
/// </summary>
/// <remarks>
/// <para>
/// This is first paragraph of remarks.
/// </para>
/// <para>
/// This is second paragraph of remarks.
/// </para>
/// </remarks>""",
            comment_code,
        )

    def test_multiple_remarks_of_mixed_lists_and_text(self) -> None:
        comment_code = self.__class__.render(
            textwrap.dedent(
                '''\
                """
                This is summary.

                This is first paragraph of remarks.

                This is a list:

                * First item
                * Second item

                This is the third paragraph.
                """

                __version__ = "dummy"
                __xml_namespace__ = "https://dummy.com"
                '''
            )
        )

        self.assertEqual(
            """\
/// <summary>
/// This is summary.
/// </summary>
/// <remarks>
/// <para>
/// This is first paragraph of remarks.
/// </para>
/// <para>
/// This is a list:
/// </para>
/// <ul>
///   <li>
///     First item
///   </li>
///   <li>
///     Second item
///   </li>
/// </ul>
/// <para>
/// This is the third paragraph.
/// </para>
/// </remarks>""",
            comment_code,
        )


if __name__ == "__main__":
    unittest.main()



"""
Copy the generated Java files for AAS v3.0 to the corresponding Java repository.

This script is practical as a part of the development pipeline:
1) Re-generate Java files (through re-recording the test data),
2) Copy the generated files (with this script),
3) Run the checks and reformatting from the Java repository, and
4) Re-generate the javadoc.
"""

import argparse
import os
import pathlib
import shutil
import sys


def main() -> int:
    """Execute the main routine."""
    repo_root = pathlib.Path(os.path.realpath(__file__)).parent.parent.parent

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "--aas_core3_java_path",
        help="Path to the aas-core3.0-java repository",
        default=str(repo_root.parent / "aas-core3.0-java"),
    )
    args = parser.parse_args()

    aas_core3_java_path = pathlib.Path(args.aas_core3_java_path)

    if not aas_core3_java_path.exists():
        print(
            f"--aas_core3_java_path does not exist: {aas_core3_java_path}",
            file=sys.stderr,
        )
        return 1

    if not aas_core3_java_path.is_dir():
        print(
            f"--aas_core3_java_path is not a directory: {aas_core3_java_path}",
            file=sys.stderr,
        )
        return 1

    expected_output_dir = (
        repo_root
        / "dev"
        / "test_data"
        / "java"
        / "test_main"
        / "aas_core_meta.v3"
        / "expected_output"
    )

    assert aas_core3_java_path.exists()

    target_src_dir = aas_core3_java_path / "src/main/java/aas_core/aas3_0"
    target_src_dir.mkdir(exist_ok=True)

    paths = sorted(expected_output_dir.glob("**/*.java"))

    print(
        f"Copying *.java\n"
        f"  from {expected_output_dir}\n"
        f"  to {target_src_dir}..."
    )
    for pth in paths:
        dst = target_src_dir / pth.relative_to(expected_output_dir)

        # NOTE (mristin):
        # We compare the contents to avoid re-compilations.
        should_copy = True
        if dst.exists():
            src_text = pth.read_text(encoding="utf-8")
            dst_text = dst.read_text(encoding="utf-8")

            should_copy = src_text != dst_text

        if should_copy:
            print(f"    Copying {pth} to {dst}...")
            shutil.copy(src=pth, dst=dst)

    print("Copied.")

    return 0


if __name__ == "__main__":
    sys.exit(main())



"""
We encountered a bug when designing V3.0. The schema constraints on
the descendant classes where not inferred if an invariant  involved properties
inherited from the parent class.

This unit test illustrates the setting, and prevents regressions.
"""


@abstract
class Abstract_lang_string(DBC):
    text: str

    def __init__(self, text: str) -> None:
        self.text = text


@invariant(
    lambda self: len(self.text) <= 128,
    "String shall have a maximum length of 128 characters.",
)
class Lang_string_name_type(Abstract_lang_string, DBC):
    def __init__(self, text: str) -> None:
        Abstract_lang_string.__init__(self, text=text)


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



@serialization(with_model_type=True)
class Something:
    some_str: str

    def __init__(self, some_str: str) -> None:
        self.some_str = some_str


class Something_more_concrete(Something):
    some_more_concrete_str: str

    def __init__(self, some_str: str, some_more_concrete_str: str) -> None:
        Something.__init__(self, some_str)

        self.some_more_concrete_str = some_more_concrete_str


class Container:
    something: Something

    def __init__(self, something: Something) -> None:
        self.something = something


__version__ = "V198.4"

__xml_namespace__ = "https://dummy/198/4"



# pylint: disable=missing-docstring

import unittest

import aas_core_codegen.java.common as java_common


class TestStringLiteral(unittest.TestCase):
    def test_empty(self) -> None:
        self.assertEqual('""', java_common.string_literal(""))

    def test_no_quotes(self) -> None:
        self.assertEqual('"x"', java_common.string_literal("x"))

    def test_tab(self) -> None:
        self.assertEqual('"a\\tb"', java_common.string_literal("a\tb"))


if __name__ == "__main__":
    unittest.main()



class Some_enum(Enum):
    """
    Represent something.

        Indented.
    """

    LITERAL = "Literal"
    """This is a literal."""

    LITERAL_WITHOUT_DOC = "LiteralWithoutDoc"


__version__ = "dummy"
__xml_namespace__ = "https://dummy.com"



@invariant(
    lambda self: len(self.name) > 2, "Constraint 1: Name with at least 3 characters"
)
class Something_concrete:
    name: str

    def __init__(self, name: str) -> None:
        self.name = name


@invariant(
    lambda self: len(self.name) < 10, "Constraint 2: Name with at most 9 characters."
)
class Something_more_concrete(Something_concrete):
    def __init__(self, name: str) -> None:
        Something_concrete.__init__(self, name)


__version__ = "V198.4"

__xml_namespace__ = "https://dummy/198/4"
